"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Table = void 0;
const dayjs_1 = __importDefault(require("dayjs"));
const lodash_1 = __importDefault(require("lodash"));
const View_1 = require("./View");
const QueryBuilder_1 = require("../../base/QueryBuilder");
const Util_1 = require("../../base/Util");
const dsl_1 = require("./dsl");
const executor_1 = require("./executor");
class Table extends View_1.View {
    /**
     * Auto convert data
     * check row data while insert or update
     * */
    checkEntity(obj, isAdd = false) {
        // checkEntity(this.schema)
        let clone = {};
        this._CONFIG.FIELD_MAP.forEach((schema, key) => {
            let field = schema.column || key;
            if (lodash_1.default.has(obj, key)) {
                clone[field] = obj[key];
            }
            let type = (0, QueryBuilder_1.getFieldType)(schema);
            if (type == 'date') {
                if (schema.isCreate) {
                    if (isAdd) {
                        clone[field] = new Date();
                    }
                    else {
                        lodash_1.default.unset(clone, field);
                    }
                    return;
                }
                if (schema.isModify) {
                    clone[field] = new Date();
                    return;
                }
            }
        });
        return clone;
    }
    /**
     * Delete documents by match {term:{filed,value}}
    */
    deleteByField(field, value) {
        const { _CONFIG: { mark, FIELD_MAP } } = this;
        if (mark)
            return this.updateByField(mark, field, value);
        let schema = FIELD_MAP.get(field);
        let column = (schema && schema.column) ? schema.column : field;
        return this.deleteByCondition([{ column, value }]);
    }
    /**
     * Delete documents by a QueryLike Schema
    */
    deleteByQuery(query) {
        const { _CONFIG: { mark } } = this;
        if (mark)
            return this.updateByQuery(mark, query);
        const condition = (0, QueryBuilder_1.queryToCondition)(query, this._CONFIG.FIELD_MAP, this._QUERY_CACHE);
        return this.deleteByCondition(condition);
    }
    deleteByCondition(condition) {
        const { _CONFIG: { mark, globalFilter }, _index } = this;
        if (mark)
            return this.updateByCondition(mark, condition);
        const param = (0, dsl_1.where)(condition);
        const query = (0, dsl_1.fixQuery)(globalFilter, param);
        return executor_1.actions.deleteByQuery(this.getClient(), _index, query);
    }
    /**
     * Delete documents by DOCUMENT._id
     *  (Not id field in Source) : Source id can use deleteByField('id',xxx);
    */
    deleteById(id) {
        const { _CONFIG: { mark } } = this;
        if (mark)
            return this.update(id, Object.assign({}, mark));
        return executor_1.actions.deleteById(this.getClient(), this._index, id);
    }
    /**
     * Update a record,  by DOCUMENT._id
     * (Not id field in Source) : Source id can use updateByField(doc, 'id',xxx);
    */
    update(id, obj, useIndex = false) {
        const entity = this.checkEntity(obj, false);
        return executor_1.actions.updateById(this.getClient(), this._index, id, entity, useIndex);
    }
    /**
     * Update documents by match {term:{filed,value}}
    */
    updateByField(obj, field, value) {
        let column = this.getColumn(field);
        return this.updateByCondition(obj, [{ column, value, fn: '=' }]);
    }
    updateByQuery(obj, query) {
        const condition = (0, QueryBuilder_1.queryToCondition)(query, this._CONFIG.FIELD_MAP, this._QUERY_CACHE);
        return this.updateByCondition(obj, condition);
    }
    updateByCondition(obj, condition) {
        const { _index, _CONFIG: { FIELD_MAP, globalFilter } } = this;
        let entity = this.checkEntity(obj);
        const scripts = [];
        for (let key of lodash_1.default.keys(entity)) {
            let column = this.getColumn(key);
            let field = this.getField(key);
            const type = (0, QueryBuilder_1.getFieldType)(FIELD_MAP.get(field));
            switch (type) {
                case 'boolean':
                    scripts.push(`ctx._source.${column}=${(0, Util_1.boolValue)(entity[key])};`);
                    break;
                case 'number':
                    scripts.push(`ctx._source.${column}=${parseFloat(entity[key])};`);
                    break;
                case 'int':
                    scripts.push(`ctx._source.${column}=${parseInt(entity[key])};`);
                    break;
                case 'date':
                    // ISO 8601
                    scripts.push(`ctx._source.${column}='${(0, dayjs_1.default)(entity[key]).format()}';`);
                    break;
                default:
                    scripts.push(`ctx._source.${column}='${entity[key]}';`);
                    break;
            }
        }
        ;
        const param = (0, dsl_1.where)(condition);
        const request = (0, dsl_1.buildSearch)(_index, param, null, null, globalFilter);
        return executor_1.actions.updateByQuery(this.getClient(), this._index, request.query, {
            source: scripts.join(''),
            lang: "painless"
        });
    }
    /**
     * Insert a record
    */
    add(object) {
        let entity = this.checkEntity(object, true);
        return executor_1.actions.add(this.getClient(), this._index, entity);
    }
}
exports.Table = Table;
