"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.buildSearch = exports.fixQuery = exports.where = exports.fixWhere = exports.SUFFIX_MATRIX = void 0;
const lodash_1 = __importDefault(require("lodash"));
const Util_1 = require("../../base/Util");
const dayjs_1 = __importDefault(require("dayjs"));
// DOCS
// https://www.elastic.co/guide/en/elasticsearch/reference/8.5/query-dsl.html
// https://www.elastic.co/guide/en/elasticsearch/painless/8.5/painless-api-reference.html
exports.SUFFIX_MATRIX = {
    'Min': { string: false, number: true, date: true, boolean: true },
    'MinThan': { string: false, number: true, date: true, boolean: false },
    'Max': { string: false, number: true, date: true, boolean: true },
    'MaxThan': { string: false, number: true, date: true, boolean: false },
    'MinH': { string: false, number: false, date: true, boolean: false },
    'MinD': { string: false, number: false, date: true, boolean: false },
    'MinM': { string: false, number: false, date: true, boolean: false },
    'MaxH': { string: false, number: false, date: true, boolean: false },
    'MaxD': { string: false, number: false, date: true, boolean: false },
    'MaxM': { string: false, number: false, date: true, boolean: false },
    'Like': { string: true, number: false, date: false, boolean: false },
    'Likel': { string: true, number: false, date: false, boolean: false },
    'Liker': { string: true, number: false, date: false, boolean: false },
    'Bt': { string: false, number: true, date: true, boolean: false },
    'BtD': { string: false, number: false, date: true, boolean: false },
    'BtY': { string: false, number: false, date: true, boolean: false },
    'BtM': { string: false, number: false, date: true, boolean: false },
    'Not': { string: true, number: true, date: true, boolean: true },
    'In': { string: true, number: true, date: false, boolean: false },
    'NotIn': { string: true, number: true, date: false, boolean: false },
    'IsNull': { string: true, number: true, date: true, boolean: true },
    'NotNull': { string: true, number: true, date: true, boolean: true },
    '>': { string: false, number: true, date: true, boolean: true },
    '>=': { string: false, number: true, date: true, boolean: false },
    '<': { string: false, number: true, date: true, boolean: true },
    '<=': { string: false, number: true, date: true, boolean: false },
    '=': { string: true, number: true, date: true, boolean: true },
    '!=': { string: true, number: true, date: true, boolean: true },
    '<>': { string: true, number: true, date: true, boolean: true },
};
const isSupport = (item, err) => {
    let suffix = exports.SUFFIX_MATRIX[item.fn] || exports.SUFFIX_MATRIX['='];
    if (suffix[item.type || 'string'])
        return true;
    err.push(`${item.column}/(${item.type}) not support method ${item.fn}`);
    return false;
};
const RANGE_MAP = new Map([
    ['Min', 'gt'],
    ['MinThan', 'gte'],
    ['Max', 'lt'],
    ['MaxThan', 'lte'],
    ['MinH', 'gte'],
    ['MinD', 'gte'],
    ['MinM', 'gte'],
    ['MaxH', 'lte'],
    ['MaxD', 'lte'],
    ['MaxM', 'lte'],
    ['>', 'gt'],
    ['>=', 'gte'],
    ['<', 'lt'],
    ['<=', 'lte'],
]);
const setRange = (query, column, value, compare) => {
    let idx = lodash_1.default.findIndex(query, pro => pro.range != undefined && pro.range[column] != undefined);
    if (idx < 0) {
        query.push({ range: { [column]: { [compare]: value } } });
    }
    else {
        query[idx].range[column][compare] = value;
    }
};
const NullCondition = (item, query, err) => {
    if (!Util_1.NONE_PARAM.has(item.fn))
        return false;
    let bool = (0, Util_1.boolValue)(item.value);
    switch (item.fn) {
        case 'IsNull':
            break;
        case 'NotNull':
            break;
    }
    if (bool) {
        query.push({ bool: { should: [{ exists: { field: item.column } }, { term: { [item.column]: null } }], minimum_should_match: 1 } });
    }
    else {
        query.push({ bool: { must_not: { exists: { field: item.column } } } });
    }
    return true;
};
const whereText = (item, query, err) => {
    switch (item.fn) {
        case 'In':
            query.push({ terms: { [item.column]: (0, Util_1.inString)(item.value + '') } });
            break;
        case 'NotIn':
            query.push({ bool: { must_not: { terms: { [item.column]: (0, Util_1.inString)(item.value + '') } } } });
            break;
        case 'Not':
            query.push({ bool: { must_not: { term: { [item.column]: item.value + '' } } } });
            break;
        case 'Like':
            query.push({ wildcard: { [item.column]: item.value + '' } });
            break;
        case 'Likel':
            query.push({ prefix: { [item.column]: item.value + '' } });
            break;
        case 'Liker':
            query.push({ wildcard: { [item.column]: '*' + item.value } });
            break;
        case '=':
            query.push({ term: { [item.column]: item.value + '' } });
            break;
        default:
            err.push(`${item.column}/(${item.type}) not support method ${item.fn}`);
            break;
    }
};
const whereNumber = (item, query, err, parseFn) => {
    let compare = RANGE_MAP.get(item.fn);
    if (compare) {
        let value = parseFn(item.value);
        setRange(query, item.column, value, compare);
        return;
    }
    // if()
    if (item.fn == 'Bt') {
        let range = (0, Util_1.betweenNumber)(item.value + '', parseFn);
        range.map(ptn => {
            let compare = RANGE_MAP.get(ptn[0]);
            if (compare) {
                setRange(query, item.column, ptn[1], compare);
            }
        });
        return;
    }
    switch (item.fn) {
        case 'In':
            query.push({ terms: { [item.column]: (0, Util_1.inNumber)(item.value + '') } });
            break;
        case 'NotIn':
            query.push({ bool: { must_not: { terms: { [item.column]: (0, Util_1.inNumber)(item.value + '') } } } });
            break;
        case '!=':
        case '<>':
            query.push({ bool: { must_not: { term: { [item.column]: item.value } } } });
            break;
        case '=':
            query.push({ term: { [item.column]: item.value } });
            break;
        default:
            err.push(`${item.column}/(${item.type}) not support method ${item.fn}`);
            break;
    }
};
const whereBoolean = (item, query, err) => {
    let bool = (0, Util_1.boolValue)(item.value);
    switch (item.fn) {
        case '<':
        case 'Min':
        case '>':
        case 'Max':
        case 'Not':
        case '!=':
        case '<>':
            query.push({ term: { [item.column]: !bool } });
            break;
        case '=':
            query.push({ term: { [item.column]: bool } });
            break;
        default:
            err.push(`${item.column}/(${item.type}) not support method ${item.fn}`);
            return;
    }
};
const whereDate = (item, query, err) => {
    if (item.value == '' || item.value == null) {
        err.push(`${item.column}/(date) : Can not be null`);
        return;
    }
    let val = null;
    if (item.fn != 'Bt') {
        val = (0, dayjs_1.default)(item.value);
        if (!val.isValid()) {
            err.push(`${item.column}/(date) : Must Be a date-string or number-stamp ${item.value}`);
            return;
        }
        switch (item.fn) {
            case 'MinH':
                val = val.startOf('hour');
                break;
            case 'MinD':
                val = val.startOf('date');
                break;
            case 'MinM':
                val = val.startOf('month');
                break;
            case 'MaxH':
                val = val.endOf('hour');
                break;
            case 'MaxD':
                val = val.endOf('date');
                break;
            case 'MaxM':
                val = val.endOf('month');
                break;
        }
    }
    const compare = RANGE_MAP.get(item.fn);
    if (compare != null) {
        setRange(query, item.column, val.format(), compare);
        return;
    }
    let start = null, end = null;
    switch (item.fn) {
        case 'BtD':
            start = val.startOf('date').format();
            end = val.endOf('date').format();
            break;
        case 'BtM':
            start = val.startOf('month').format();
            end = val.endOf('month').format();
            break;
        case 'BtY':
            start = val.startOf('year').format();
            end = val.endOf('year').format();
            break;
        case 'Bt':
            let range = (0, Util_1.betweenDate)(item.value + '');
            range.map(ptn => {
                let compare = RANGE_MAP.get(ptn[0]);
                if (compare) {
                    setRange(query, item.column, ptn[1], compare);
                }
            });
            return;
        case '!=':
        case '<>':
        case 'Not':
            query.push({ bool: { must_not: { term: { [item.column]: val.toDate() } } } });
            return;
        case '=':
            query.push({ term: { [item.column]: val.toDate() } });
            return;
        default:
            return;
    }
    setRange(query, item.column, start, 'lte');
    setRange(query, item.column, end, 'gte');
};
const ItemToWhere = (whereItem, query, err) => {
    let item = Object.assign(Object.assign({}, whereItem), { fn: whereItem.fn ? whereItem.fn : '=', type: whereItem.type ? whereItem.type : 'string' });
    if (!isSupport(item, err))
        return;
    if (NullCondition(whereItem, query, err))
        return;
    switch (item.type) {
        case 'number':
        case 'int':
            whereNumber(item, query, err, item.type == 'int' ? parseInt : parseFloat);
            return;
        case 'string':
            whereText(item, query, err);
            return;
        case 'boolean':
            whereBoolean(item, query, err);
            return;
        case 'date':
            whereDate(item, query, err);
            return;
        default:
            whereText(item, query, err);
            return;
    }
};
const ConditionToWhere = (root, query, err) => {
    for (let group of root.items) {
        if (lodash_1.default.has(group, 'link')) {
            let children = [];
            ConditionToWhere(group, children, err);
            if (children.length == 0) {
                continue;
            }
            if (children.length == 1) {
                query.push(children[0]);
            }
            if (root.link == 'OR') {
                query.push({ bool: { should: children, minimum_should_match: 1 } });
            }
            query.push({ bool: { filter: children } });
        }
        else {
            ItemToWhere(group, query, err);
        }
    }
};
const buildWhere = (condition) => {
    let root = lodash_1.default.isArray(condition) ? { link: 'AND', items: condition } : condition;
    let query = [];
    let err = [];
    ConditionToWhere(root, query, err);
    (0, Util_1.throwErr)(err, 'Some DSL Error Occur');
    return query;
};
const fixWhere = (condition) => {
    let query = buildWhere(condition);
    if (query.length == 0) {
        return null;
    }
    let bool = { filter: query };
    return { constant_score: { filter: { bool } } };
};
exports.fixWhere = fixWhere;
const where = (condition) => {
    let query = buildWhere(condition);
    if (query.length == 0) {
        return null;
    }
    if (query.length == 1) {
        return { constant_score: { filter: query[0] } };
    }
    if (!lodash_1.default.isArray(condition) && condition.link == 'OR') {
        return { constant_score: { filter: { bool: { should: query, minimum_should_match: 1 } } } };
    }
    return { constant_score: { filter: { bool: { filter: query } } } };
};
exports.where = where;
/**
 * Global Filter Must postion on first Arg
*/
const fixQuery = (fixFilter, query) => {
    if (fixFilter == null && query == null) {
        return { match_all: {} };
    }
    if (fixFilter == null) {
        return query;
    }
    if (query == null) {
        return fixFilter;
    }
    if (!lodash_1.default.has(query, 'constant_score')) {
        return lodash_1.default.merge({}, fixFilter, query);
    }
    let searchParam = lodash_1.default.cloneDeep(fixFilter);
    lodash_1.default.keys(query.constant_score.filter).map(key => {
        if (key == 'bool') {
            lodash_1.default.keys(query.constant_score.filter.bool).map(boolKey => {
                if (boolKey == 'filter' || boolKey == 'must') {
                    if (lodash_1.default.isArray(query.constant_score.filter.bool[boolKey])) {
                        query.constant_score.filter.bool[boolKey].map(item => {
                            searchParam.constant_score.filter.bool.filter.push(item);
                        });
                        // (searchParam.constant_score.filter.bool.filter as QueryDslQueryContainer[]).concat(query.constant_score.filter.bool[boolKey] as QueryDslQueryContainer[])
                    }
                    else {
                        searchParam.constant_score.filter.bool.filter.push(query.constant_score.filter.bool[boolKey]);
                    }
                }
                else {
                    searchParam.constant_score.filter.bool[boolKey] = query.constant_score.filter.bool[boolKey];
                }
            });
        }
        else {
            searchParam.constant_score.filter.bool.filter.push({ [key]: query.constant_score.filter[key] });
        }
    });
    return searchParam;
};
exports.fixQuery = fixQuery;
const buildSearch = (indexName, query, order, _source_excludes, globalFilter) => {
    const param = Object.assign({ index: indexName }, order);
    if (_source_excludes && _source_excludes.length) {
        param._source_excludes = _source_excludes;
    }
    param.query = (0, exports.fixQuery)(globalFilter, query);
    return param;
};
exports.buildSearch = buildSearch;
