"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.fixWhere = exports.where = exports.SUFFIX_MATRIX = void 0;
const lodash_1 = __importDefault(require("lodash"));
const typebox_1 = require("@sinclair/typebox");
const Util_1 = require("../../base/Util");
const dayjs_1 = __importDefault(require("dayjs"));
// https://www.postgresql.org/docs/15/queries-table-expressions.html
exports.SUFFIX_MATRIX = {
    'Min': { string: true, number: true, date: true, boolean: true },
    'MinThan': { string: true, number: true, date: true, boolean: false },
    'Max': { string: true, number: true, date: true, boolean: true },
    'MaxThan': { string: true, number: true, date: true, boolean: false },
    'MinH': { string: false, number: false, date: true, boolean: false },
    'MinD': { string: false, number: false, date: true, boolean: false },
    'MinM': { string: false, number: false, date: true, boolean: false },
    'MaxH': { string: false, number: false, date: true, boolean: false },
    'MaxD': { string: false, number: false, date: true, boolean: false },
    'MaxM': { string: false, number: false, date: true, boolean: false },
    'Like': { string: true, number: false, date: false, boolean: false },
    'Likel': { string: true, number: false, date: false, boolean: false },
    'Liker': { string: true, number: false, date: false, boolean: false },
    'Bt': { string: false, number: true, date: true, boolean: false },
    'BtD': { string: false, number: false, date: true, boolean: false },
    'BtY': { string: false, number: false, date: true, boolean: false },
    'BtM': { string: false, number: false, date: true, boolean: false },
    'Not': { string: true, number: true, date: true, boolean: true },
    'In': { string: true, number: true, date: false, boolean: false },
    'NotIn': { string: true, number: true, date: false, boolean: false },
    'IsNull': { string: true, number: true, date: true, boolean: true },
    'NotNull': { string: true, number: true, date: true, boolean: true },
    // 'IsDistinct': { string: true, number: true, date: false, boolean: false },
    // 'NotDistinct': { string: true, number: true, date: false, boolean: false },
    '>': { string: true, number: true, date: true, boolean: true },
    '>=': { string: true, number: true, date: true, boolean: false },
    '<': { string: true, number: true, date: true, boolean: true },
    '<=': { string: true, number: true, date: true, boolean: false },
    '=': { string: true, number: true, date: true, boolean: true },
    '!=': { string: true, number: true, date: true, boolean: true },
    '<>': { string: true, number: true, date: true, boolean: true },
};
const isSupport = (item, err) => {
    let suffix = exports.SUFFIX_MATRIX[item.fn] || exports.SUFFIX_MATRIX['='];
    if (suffix[item.type || 'string'])
        return true;
    err.push(`${item.column}/(${item.type}) not support method ${item.fn}`);
    return false;
};
const NullCondition = (item, pos) => {
    if (!Util_1.NONE_PARAM.has(item.fn))
        return false;
    let bool = (0, Util_1.boolValue)(item.value);
    switch (item.fn) {
        case 'IsNull':
            pos.SQL.push(`${item.column} IS ${bool ? '' : 'NOT'} NULL`);
            return true;
        case 'NotNull':
            pos.SQL.push(`${item.column} IS ${bool ? 'NOT' : ''} NULL`);
            return true;
        // case 'IsDistinct':
        //     pos.SQL.push(`${item.column} IS ${bool ? '' : 'NOT'} DISTINCT`);
        //     return true;
        // case 'NotDistinct':
        //     pos.SQL.push(`${item.column} IS ${bool ? 'NOT' : ''} DISTINCT`);
        //     return true;
    }
};
const compareSuffix = (suffix) => {
    switch (suffix) {
        case '<':
        case '>':
        case '<=':
        case '>=':
        case '=':
            return suffix;
        case 'Min':
            return '>';
        case 'Max':
            return '<';
        case 'MinD':
        case 'MinH':
        case 'MinM':
        case 'MinThan':
            return '>=';
        case 'MaxD':
        case 'MaxH':
        case 'MaxM':
        case 'MaxThan':
            return '<=';
        case 'Not':
        case '!=':
        case '<>':
            return '!=';
    }
    return null;
};
const whereText = (item, pos, err) => {
    const compare = compareSuffix(item.fn);
    if (compare != null) {
        pos.SQL.push(`${item.column} ${compare} $${pos.NUM}`);
        pos.PARAM.push(item.value);
        pos.NUM++;
        return;
    }
    if (InFn.has(item.fn)) {
        pos.SQL.push(`${item.column} ${InFn.get(item.fn)} ANY($${pos.NUM}::text[])`);
        pos.PARAM.push((0, Util_1.inString)(item.value + ''));
        pos.NUM++;
        return;
    }
    switch (item.fn) {
        case 'Like':
            pos.SQL.push(`${item.column} LIKE $${pos.NUM}`);
            pos.PARAM.push('%' + item.value + '%');
            pos.NUM++;
            return;
        case 'Likel':
            pos.SQL.push(`${item.column} LIKE $${pos.NUM}`);
            pos.PARAM.push(item.value + '%');
            pos.NUM++;
            return;
        case 'Liker':
            pos.SQL.push(`${item.column} LIKE $${pos.NUM}`);
            pos.PARAM.push('%' + item.value);
            pos.NUM++;
            return;
        default:
            err.push(`${item.column}/ type : String not support method ${item.fn}`);
            return;
    }
};
const InFn = new Map([['In', '='], ['NotIn', '!=']]);
const whereNumber = (item, pos, err, parseFn) => {
    const compare = compareSuffix(item.fn);
    if (compare != null) {
        try {
            let val = lodash_1.default.isNumber(item.value) ? item.value : parseFn(item.value);
            pos.SQL.push(`${item.column} ${compare} $${pos.NUM}`);
            pos.PARAM.push(val);
            pos.NUM++;
        }
        catch (_a) {
            err.push(`${item.column}/ type : Number value is not a Number ${item.value}`);
        }
        return;
    }
    if (InFn.has(item.fn)) {
        pos.SQL.push(`${item.column} ${InFn.get(item.fn)} ANY($${pos.NUM}::int[])`);
        pos.PARAM.push((0, Util_1.inNumber)(item.value + ''));
        pos.NUM++;
        return;
    }
    if (item.fn == 'Bt') {
        let range = (0, Util_1.betweenNumber)(item.value + '', parseFn);
        if (range == null) {
            err.push(`${item.column}/(Number) :  Between Value invalidated ${item.value}`);
            return;
        }
        range.map(oper => {
            pos.SQL.push(`${item.column} ${oper[0]} $${pos.NUM}`);
            pos.PARAM.push(oper[1]);
            pos.NUM++;
        });
        return;
    }
    err.push(`${item.column}/(Number) : not support method ${item.fn}`);
};
const whereDate = (item, pos, err) => {
    let val = null;
    if (item.fn != 'Bt') {
        if (item.value == '' || item.value == null) {
            err.push(`${item.column}/(date) : Can not be null`);
            return;
        }
        val = (0, dayjs_1.default)(item.value);
        if (!val.isValid()) {
            err.push(`${item.column}/(date) : Must Be a date-string or number-stamp ${item.value}`);
            return;
        }
        switch (item.fn) {
            case 'MinH':
                val = val.startOf('hour');
                break;
            case 'MinD':
                val = val.startOf('date');
                break;
            case 'MinM':
                val = val.startOf('month');
                break;
            case 'MaxH':
                val = val.endOf('hour');
                break;
            case 'MaxD':
                val = val.endOf('date');
                break;
            case 'MaxM':
                val = val.endOf('month');
                break;
        }
    }
    const compare = compareSuffix(item.fn);
    if (compare != null) {
        pos.SQL.push(`${item.column} ${compare} $${pos.NUM}`);
        pos.PARAM.push(val.toDate());
        pos.NUM++;
        return;
    }
    let start = null, end = null;
    switch (item.fn) {
        case 'BtD':
            start = val.startOf('date').toDate();
            end = val.endOf('date').toDate();
            break;
        case 'BtM':
            start = val.startOf('month').toDate();
            end = val.endOf('month').toDate();
            break;
        case 'BtY':
            start = val.startOf('year').toDate();
            end = val.endOf('year').toDate();
            break;
        case 'Bt':
            const range = (0, Util_1.betweenDate)(item.value + '');
            if (range == null) {
                err.push(`${item.column}/(Date) :  Between Value invalidated ${item.value}`);
                return;
            }
            range.map(oper => {
                pos.SQL.push(`${item.column} ${oper[0]} $${pos.NUM}`);
                pos.PARAM.push(oper[1]);
                pos.NUM++;
            });
            return;
    }
    if (start == null || end == null) {
        err.push(`${item.column}/(Date) : not support method ${item.fn}`);
        return;
    }
    pos.SQL.push(`${item.column} >= $${pos.NUM}`);
    pos.PARAM.push(start);
    pos.NUM++;
    pos.SQL.push(`${item.column} <= $${pos.NUM}`);
    pos.PARAM.push(end);
    pos.NUM++;
};
const whereBoolean = (item, pos, err) => {
    let bool = (0, Util_1.boolValue)(item.value);
    switch (item.fn) {
        case 'IsNull':
            pos.SQL.push(`${item.column} IS ${bool ? '' : 'NOT'} NULL`);
            return;
        case 'NotNull':
            pos.SQL.push(`${item.column} IS ${bool ? 'NOT' : ''} NULL`);
            return;
        case '!=':
        case '<>':
        case 'Not':
        case 'Min':
        case 'Max':
        case '>':
        case '<':
            bool = !bool;
            break;
        default:
            break;
    }
    pos.SQL.push(`${item.column} IS ${bool ? '' : 'NOT'} TRUE`);
};
const ItemToWhere = (whereItem, pos, err) => {
    let item = Object.assign(Object.assign({}, whereItem), { fn: whereItem.fn ? whereItem.fn : '=', type: whereItem.type ? whereItem.type : 'string' });
    if (!isSupport(item, err))
        return;
    if (NullCondition(item, pos))
        return;
    switch (item.type) {
        case 'number':
        case 'int':
            whereNumber(item, pos, err, item.type == 'int' ? parseInt : parseFloat);
            return;
        case 'string':
            whereText(item, pos, err);
            return;
        case 'boolean':
            whereBoolean(item, pos, err);
            return;
        case 'date':
            whereDate(item, pos, err);
            return;
        default:
            whereText(item, pos, err);
            return;
    }
};
const ConditionToWhere = (condition, pos, err) => {
    for (let item of condition.items) {
        if (lodash_1.default.has(item, 'link')) {
            let group = item;
            let _pos = {
                SQL: [],
                PARAM: [],
                NUM: pos.NUM
            };
            ConditionToWhere(group, _pos, err);
            if (_pos.SQL.length) {
                if (_pos.NUM > pos.NUM)
                    pos.NUM = _pos.NUM;
                for (let param of _pos.PARAM)
                    pos.PARAM.push(param);
                // let link = group.link == 'NOT' ? 'AND NOT' : group.link;
                pos.SQL.push('(' + _pos.SQL.join(' ' + group.link + ' ') + ')');
            }
        }
        else {
            ItemToWhere(item, pos, err);
        }
    }
};
const where = (condition, startIdx = 1) => {
    const pos = { SQL: [], PARAM: [], NUM: startIdx };
    let root = lodash_1.default.isArray(condition) ? { link: 'AND', items: condition } : condition;
    let err = [];
    ConditionToWhere(root, pos, err);
    (0, Util_1.throwErr)(err, 'Some SQL Error Occur');
    if (pos.SQL.length == 0) {
        return ['', []];
    }
    ;
    return [pos.SQL.join(" " + root.link + " "), pos.PARAM];
};
exports.where = where;
const fixWhere = (FIELD_MAP, extra) => {
    let ITEMS = [];
    let ctf = new Map();
    const convert = (kind, value) => {
        switch (kind) {
            case 'Boolean':
                return value;
            case 'Number':
                return parseFloat(value);
            case 'Integer':
                return parseInt(value);
            default:
                return value + '';
        }
    };
    for (let [key, val] of FIELD_MAP) {
        if (lodash_1.default.has(val, 'delMark')) {
            ITEMS.push({ column: (val.column || key), fn: '<>', value: convert(val[typebox_1.Kind], val.delMark) });
        }
        ctf.set((val.column || key), key);
    }
    extra.map(item => {
        // inner usage field
        // @ts-ignore
        let schema = FIELD_MAP.get(item.field) || FIELD_MAP.get(ctf.get(item.field));
        if (schema == null)
            return;
        ITEMS.push(Object.assign(Object.assign({}, item), { value: convert(schema[typebox_1.Kind], item.value) }));
    });
    if (ITEMS.length == 0)
        return ['', ' WHERE '];
    let [SQL, PARAM] = (0, exports.where)(ITEMS);
    if (SQL.length == 0)
        return ['', ' WHERE '];
    PARAM.map((item, i) => {
        SQL = SQL.replaceAll(`$${i + 1}`, lodash_1.default.isNumber(item) ? (item + '') : `'${item}'`);
    });
    return [' WHERE ' + SQL, ' AND '];
};
exports.fixWhere = fixWhere;
